package com.sun.xml.internal.bind.v2.schemagen;

import com.sun.xml.internal.bind.v2.schemagen.xmlschema.TopLevelElement;
import com.sun.xml.internal.bind.v2.schemagen.xmlschema.ExplicitGroup;
import com.sun.xml.internal.bind.v2.schemagen.xmlschema.Any;
import com.sun.xml.internal.bind.v2.schemagen.xmlschema.AttributeType;
import com.sun.xml.internal.bind.v2.schemagen.xmlschema.LocalAttribute;
import java.util.Collection;
import com.sun.xml.internal.bind.v2.model.impl.ClassInfoImpl;
import com.sun.xml.internal.bind.v2.schemagen.xmlschema.ContentModelContainer;
import com.sun.xml.internal.bind.v2.model.core.MapPropertyInfo;
import com.sun.xml.internal.txw2.TypedXmlWriter;
import com.sun.xml.internal.bind.v2.schemagen.xmlschema.SimpleExtension;
import com.sun.xml.internal.bind.v2.schemagen.xmlschema.ComplexExtension;
import com.sun.xml.internal.bind.v2.schemagen.xmlschema.TypeDefParticle;
import com.sun.xml.internal.bind.v2.schemagen.xmlschema.AttrDecls;
import com.sun.xml.internal.bind.v2.model.core.ValuePropertyInfo;
import com.sun.xml.internal.bind.v2.TODO;
import com.sun.xml.internal.bind.v2.model.core.ReferencePropertyInfo;
import java.util.ArrayList;
import com.sun.xml.internal.bind.v2.schemagen.xmlschema.ComplexTypeHost;
import com.sun.xml.internal.bind.v2.schemagen.xmlschema.SimpleRestrictionModel;
import com.sun.xml.internal.bind.v2.schemagen.xmlschema.SimpleType;
import com.sun.xml.internal.bind.v2.model.core.EnumConstant;
import com.sun.xml.internal.bind.v2.schemagen.xmlschema.LocalElement;
import com.sun.xml.internal.bind.v2.schemagen.xmlschema.ComplexType;
import org.xml.sax.SAXParseException;
import com.sun.xml.internal.bind.v2.schemagen.xmlschema.SimpleTypeHost;
import com.sun.xml.internal.bind.v2.model.core.MaybeElement;
import com.sun.xml.internal.bind.v2.schemagen.xmlschema.TypeHost;
import com.sun.xml.internal.bind.v2.schemagen.xmlschema.TopLevelAttribute;
import com.sun.xml.internal.bind.v2.schemagen.xmlschema.Import;
import com.sun.xml.internal.txw2.TxwException;
import com.sun.xml.internal.txw2.output.ResultFactory;
import com.sun.xml.internal.bind.v2.schemagen.xmlschema.Schema;
import com.sun.istack.internal.Nullable;
import java.util.List;
import com.sun.xml.internal.bind.v2.model.core.Element;
import com.sun.xml.internal.bind.v2.model.core.TypeInfo;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.Set;
import com.sun.istack.internal.NotNull;
import java.io.File;
import java.net.URISyntaxException;
import java.net.URI;
import com.sun.xml.internal.bind.v2.model.core.WildcardMode;
import com.sun.xml.internal.bind.v2.model.core.Adapter;
import com.sun.xml.internal.bind.v2.runtime.SwaRefAdapter;
import com.sun.xml.internal.bind.v2.model.core.NonElementRef;
import java.io.IOException;
import java.io.Writer;
import java.io.OutputStream;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.Result;
import java.util.HashMap;
import com.sun.xml.internal.bind.v2.util.StackRecorder;
import java.util.logging.Level;
import javax.xml.bind.SchemaOutputResolver;
import com.sun.xml.internal.txw2.TXW;
import com.sun.xml.internal.bind.v2.schemagen.episode.Bindings;
import com.sun.xml.internal.txw2.output.XmlSerializer;
import javax.xml.bind.annotation.XmlElement;
import javax.activation.MimeType;
import javax.xml.namespace.QName;
import com.sun.xml.internal.bind.v2.model.core.TypeRef;
import com.sun.xml.internal.bind.v2.model.core.ElementPropertyInfo;
import com.sun.xml.internal.bind.v2.model.core.AttributePropertyInfo;
import com.sun.xml.internal.bind.v2.model.core.PropertyInfo;
import com.sun.xml.internal.bind.api.CompositeStructure;
import java.util.Iterator;
import com.sun.xml.internal.bind.v2.model.core.ArrayInfo;
import com.sun.xml.internal.bind.v2.model.core.EnumLeafInfo;
import com.sun.xml.internal.bind.v2.model.core.ElementInfo;
import java.util.TreeMap;
import java.util.Comparator;
import com.sun.xml.internal.bind.v2.model.core.ClassInfo;
import com.sun.xml.internal.bind.v2.util.CollisionCheckStack;
import com.sun.xml.internal.bind.v2.model.core.NonElement;
import com.sun.xml.internal.bind.v2.model.core.TypeInfoSet;
import com.sun.xml.internal.bind.v2.model.nav.Navigator;
import com.sun.xml.internal.bind.api.ErrorListener;
import java.util.Map;
import java.util.logging.Logger;

public final class XmlSchemaGenerator<T, C, F, M>
{
    private static final Logger logger;
    private final Map<String, XmlSchemaGenerator.Namespace> namespaces;
    private ErrorListener errorListener;
    private Navigator<T, C, F, M> navigator;
    private final TypeInfoSet<T, C, F, M> types;
    private final NonElement<T, C> stringType;
    private final NonElement<T, C> anyType;
    private final CollisionCheckStack<ClassInfo<T, C>> collisionChecker;
    private static final Comparator<String> NAMESPACE_COMPARATOR;
    private static final String newline = "\n";
    
    public XmlSchemaGenerator(Navigator<T, C, F, M> navigator, TypeInfoSet<T, C, F, M> types) {
        this.namespaces = new TreeMap(XmlSchemaGenerator.NAMESPACE_COMPARATOR);
        this.collisionChecker = new CollisionCheckStack();
        this.navigator = navigator;
        this.types = types;
        this.stringType = types.getTypeInfo(navigator.ref(String.class));
        this.anyType = types.getAnyTypeInfo();
        for (ClassInfo<T, C> ci : types.beans().values()) {
            this.add(ci);
        }
        for (ElementInfo<T, C> ei1 : types.getElementMappings(null).values()) {
            this.add(ei1);
        }
        for (EnumLeafInfo<T, C> ei2 : types.enums().values()) {
            this.add(ei2);
        }
        for (ArrayInfo<T, C> a : types.arrays().values()) {
            this.add(a);
        }
    }
    
    private XmlSchemaGenerator.Namespace getNamespace(String uri) {
        XmlSchemaGenerator.Namespace n = (XmlSchemaGenerator.Namespace)this.namespaces.get(uri);
        if (n == null) {
            this.namespaces.put(uri, n = new XmlSchemaGenerator.Namespace(uri));
            return n;
        }
        return n;
    }
    
    public void add(ClassInfo<T, C> clazz) {
        assert clazz != null;
        String nsUri = null;
        if (clazz.getClazz() == this.navigator.asDecl(CompositeStructure.class)) {
            return;
        }
        if (clazz.isElement()) {
            nsUri = clazz.getElementName().getNamespaceURI();
            XmlSchemaGenerator.Namespace ns = this.getNamespace(nsUri);
            ns.classes.add(clazz);
            XmlSchemaGenerator.Namespace.access$100(ns, clazz.getTypeName());
            this.add(clazz.getElementName(), false, clazz);
        }
        QName tn = clazz.getTypeName();
        if (tn != null) {
            nsUri = tn.getNamespaceURI();
        }
        else if (nsUri == null) {
            return;
        }
        XmlSchemaGenerator.Namespace n = this.getNamespace(nsUri);
        n.classes.add(clazz);
        for (PropertyInfo<T, C> p : clazz.getProperties()) {
            XmlSchemaGenerator.Namespace.access$200(n, p, 1);
            if ((p instanceof AttributePropertyInfo)) {
                AttributePropertyInfo<T, C> ap = (AttributePropertyInfo)p;
                String aUri = ap.getXmlName().getNamespaceURI();
                if (aUri.length() > 0) {
                    this.getNamespace(aUri).addGlobalAttribute(ap);
                    XmlSchemaGenerator.Namespace.access$100(n, ap.getXmlName());
                }
            }
            if ((p instanceof ElementPropertyInfo)) {
                ElementPropertyInfo<T, C> ep = (ElementPropertyInfo)p;
                for (TypeRef<T, C> tref : ep.getTypes()) {
                    String eUri = tref.getTagName().getNamespaceURI();
                    if (eUri.length() > 0 && !eUri.equals(n.uri)) {
                        this.getNamespace(eUri).addGlobalElement(tref);
                        XmlSchemaGenerator.Namespace.access$100(n, tref.getTagName());
                    }
                    else {
                        continue;
                    }
                }
            }
            if (this.generateSwaRefAdapter(p)) {
                n.useSwaRef = true;
            }
            MimeType mimeType = p.getExpectedMimeType();
            if (mimeType != null) {
                n.useMimeNs = true;
            }
            else {
                continue;
            }
        }
        ClassInfo<T, C> bc = clazz.getBaseClass();
        if (bc != null) {
            this.add(bc);
            XmlSchemaGenerator.Namespace.access$100(n, bc.getTypeName());
        }
        return;
    }
    
    public void add(ElementInfo<T, C> elem) {
        assert elem != null;
        boolean nillable = false;
        QName name = elem.getElementName();
        XmlSchemaGenerator.Namespace n = this.getNamespace(name.getNamespaceURI());
        ElementInfo ei;
        if (elem.getScope() != null) {
            ei = this.types.getElementInfo(elem.getScope().getClazz(), name);
        }
        else {
            ei = this.types.getElementInfo(null, name);
        }
        XmlElement xmlElem = (XmlElement)ei.getProperty().readAnnotation(XmlElement.class);
        nillable = (xmlElem != null && xmlElem.nillable());
        n.elementDecls.put(name.getLocalPart(), new Namespace.ElementWithType(nillable, elem.getContentType()));
        XmlSchemaGenerator.Namespace.access$200(n, elem.getProperty(), 1);
        return;
    }
    
    public void add(EnumLeafInfo<T, C> envm) {
        assert envm != null;
        String nsUri = null;
        if (envm.isElement()) {
            nsUri = envm.getElementName().getNamespaceURI();
            XmlSchemaGenerator.Namespace ns = this.getNamespace(nsUri);
            ns.enums.add(envm);
            XmlSchemaGenerator.Namespace.access$100(ns, envm.getTypeName());
            this.add(envm.getElementName(), false, envm);
        }
        QName typeName = envm.getTypeName();
        if (typeName != null) {
            nsUri = typeName.getNamespaceURI();
        }
        else if (nsUri == null) {
            return;
        }
        XmlSchemaGenerator.Namespace n = this.getNamespace(nsUri);
        n.enums.add(envm);
        XmlSchemaGenerator.Namespace.access$100(n, envm.getBaseType().getTypeName());
        return;
    }
    
    public void add(ArrayInfo<T, C> a) {
        assert a != null;
        String namespaceURI = a.getTypeName().getNamespaceURI();
        XmlSchemaGenerator.Namespace n = this.getNamespace(namespaceURI);
        n.arrays.add(a);
        XmlSchemaGenerator.Namespace.access$100(n, a.getItemType().getTypeName());
        return;
    }
    
    public void add(QName tagName, boolean isNillable, NonElement<T, C> type) {
        if (type != null && type.getType() == this.navigator.ref(CompositeStructure.class)) {
            return;
        }
        XmlSchemaGenerator.Namespace n = this.getNamespace(tagName.getNamespaceURI());
        n.elementDecls.put(tagName.getLocalPart(), new Namespace.ElementWithType(isNillable, type));
        if (type != null) {
            XmlSchemaGenerator.Namespace.access$100(n, type.getTypeName());
        }
        return;
    }
    
    public void writeEpisodeFile(XmlSerializer out) {
        Bindings root = (Bindings)TXW.create(Bindings.class, out);
        if (this.namespaces.containsKey("")) {
            root._namespace("http://java.sun.com/xml/ns/jaxb", "jaxb");
        }
        root.version("2.1");
        for (Map.Entry<String, XmlSchemaGenerator.Namespace> e : this.namespaces.entrySet()) {
            Bindings group = root.bindings();
            String tns = (String)e.getKey();
            String prefix;
            if (!tns.equals("")) {
                group._namespace(tns, "tns");
                prefix = "tns:";
            }
            else {
                prefix = "";
            }
            group.scd("x-schema::" + (tns.equals("") ? "" : "tns"));
            group.schemaBindings().map(false);
            for (ClassInfo<T, C> ci : ((XmlSchemaGenerator.Namespace)e.getValue()).classes) {
                if (ci.getTypeName() == null) {
                    continue;
                }
                if (ci.getTypeName().getNamespaceURI().equals(tns)) {
                    Bindings child = group.bindings();
                    child.scd(126 + prefix + ci.getTypeName().getLocalPart());
                    child.klass().ref(ci.getName());
                }
                if (ci.isElement() && ci.getElementName().getNamespaceURI().equals(tns)) {
                    Bindings child = group.bindings();
                    child.scd(prefix + ci.getElementName().getLocalPart());
                    child.klass().ref(ci.getName());
                }
                else {
                    continue;
                }
            }
            for (EnumLeafInfo<T, C> en : ((XmlSchemaGenerator.Namespace)e.getValue()).enums) {
                if (en.getTypeName() == null) {
                    continue;
                }
                Bindings child = group.bindings();
                child.scd(126 + prefix + en.getTypeName().getLocalPart());
                child.klass().ref(this.navigator.getClassName(en.getClazz()));
            }
            group.commit(true);
        }
        root.commit();
    }
    
    public void write(SchemaOutputResolver resolver, ErrorListener errorListener) throws IOException {
        if (resolver == null) {
            throw new IllegalArgumentException();
        }
        if (XmlSchemaGenerator.logger.isLoggable(Level.FINE)) {
            XmlSchemaGenerator.logger.log(Level.FINE, "Writing XML Schema for " + this.toString(), new StackRecorder());
        }
        resolver = new FoolProofResolver(resolver);
        this.errorListener = errorListener;
        Map<String, String> schemaLocations = this.types.getSchemaLocations();
        Map<XmlSchemaGenerator.Namespace, Result> out = new HashMap();
        Map<XmlSchemaGenerator.Namespace, String> systemIds = new HashMap();
        this.namespaces.remove("http://www.w3.org/2001/XMLSchema");
        for (XmlSchemaGenerator.Namespace n : this.namespaces.values()) {
            String schemaLocation = (String)schemaLocations.get(n.uri);
            if (schemaLocation != null) {
                systemIds.put(n, schemaLocation);
            }
            else {
                Result output = resolver.createOutput(n.uri, "schema" + (out.size() + 1) + ".xsd");
                if (output != null) {
                    out.put(n, output);
                    systemIds.put(n, output.getSystemId());
                }
            }
            n.resetWritten();
        }
        for (Map.Entry<XmlSchemaGenerator.Namespace, Result> e : out.entrySet()) {
            Result result = (Result)e.getValue();
            XmlSchemaGenerator.Namespace.access$800((XmlSchemaGenerator.Namespace)e.getKey(), result, systemIds);
            if ((result instanceof StreamResult)) {
                OutputStream outputStream = ((StreamResult)result).getOutputStream();
                if (outputStream != null) {
                    outputStream.close();
                }
                else {
                    Writer writer = ((StreamResult)result).getWriter();
                    if (writer != null) {
                        writer.close();
                    }
                    else {
                        continue;
                    }
                }
            }
            else {
                continue;
            }
        }
        return;
    }
    
    private boolean generateSwaRefAdapter(NonElementRef<T, C> typeRef) {
        return this.generateSwaRefAdapter(typeRef.getSource());
    }
    
    private boolean generateSwaRefAdapter(PropertyInfo<T, C> prop) {
        Adapter<T, C> adapter = prop.getAdapter();
        if (adapter == null) {
            return false;
        }
        Object o = this.navigator.asDecl(SwaRefAdapter.class);
        return o != null && o.equals(adapter.adapterType);
    }
    
    @Override
    public String toString() {
        StringBuilder buf = new StringBuilder();
        for (XmlSchemaGenerator.Namespace ns : this.namespaces.values()) {
            if (buf.length() > 0) {
                buf.append(',');
            }
            buf.append(ns.uri).append('=').append(ns);
        }
        return super.toString() + 91 + buf + 93;
    }
    
    private static String getProcessContentsModeName(WildcardMode wc) {
        switch (wc) {
            case LAX:
            case SKIP: {
                return wc.name().toLowerCase();
            }
            case STRICT: {
                return null;
            }
            default: {
                throw new IllegalStateException();
            }
        }
    }
    
    protected static String relativize(String uri, String baseUri) {
        try {
            assert uri != null;
            if (baseUri == null) {
                return uri;
            }
            URI theUri = new URI(Util.escapeURI(uri));
            URI theBaseUri = new URI(Util.escapeURI(baseUri));
            if (theUri.isOpaque() || theBaseUri.isOpaque()) {
                return uri;
            }
            if (!Util.equalsIgnoreCase(theUri.getScheme(), theBaseUri.getScheme()) || !Util.equal(theUri.getAuthority(), theBaseUri.getAuthority())) {
                return uri;
            }
            String uriPath = theUri.getPath();
            String basePath = theBaseUri.getPath();
            if (!basePath.endsWith("/")) {
                basePath = Util.normalizeUriPath(basePath);
            }
            if (uriPath.equals(basePath)) {
                return ".";
            }
            String relPath = calculateRelativePath(uriPath, basePath, fixNull(theUri.getScheme()).equals("file"));
            if (relPath == null) {
                return uri;
            }
            StringBuilder relUri = new StringBuilder();
            relUri.append(relPath);
            if (theUri.getQuery() != null) {
                relUri.append('?').append(theUri.getQuery());
            }
            if (theUri.getFragment() != null) {
                relUri.append('#').append(theUri.getFragment());
            }
            return relUri.toString();
        }
        catch (URISyntaxException e) {
            throw new InternalError("Error escaping one of these uris:\n\t" + uri + "\n\t" + baseUri);
        }
    }
    
    private static String fixNull(String s) {
        if (s == null) {
            return "";
        }
        return s;
    }
    
    private static String calculateRelativePath(String uri, String base, boolean fileUrl) {
        boolean onWindows = File.pathSeparatorChar == ';';
        if (base == null) {
            return null;
        }
        if ((fileUrl && onWindows && startsWithIgnoreCase(uri, base)) || uri.startsWith(base)) {
            return uri.substring(base.length());
        }
        return "../" + calculateRelativePath(uri, Util.getParentUriPath(base), fileUrl);
    }
    
    private static boolean startsWithIgnoreCase(String s, String t) {
        return s.toUpperCase().startsWith(t.toUpperCase());
    }
    
    static XmlSchemaGenerator.Namespace access$1100(XmlSchemaGenerator x0, String x1) {
        return x0.getNamespace(x1);
    }
    
    static boolean access$1500(XmlSchemaGenerator x0, NonElementRef x1) {
        return x0.generateSwaRefAdapter(x1);
    }
    
    static String access$2200(WildcardMode x0) {
        return getProcessContentsModeName(x0);
    }
    
    static {
        logger = com.sun.xml.internal.bind.Util.getClassLogger();
        NAMESPACE_COMPARATOR = new Comparator<String>() {
            @Override
            public int compare(String lhs, String rhs) {
                return -lhs.compareTo(rhs);
            }
        };
    }
    
    private class Namespace
    {
        @NotNull
        final String uri;
        private final Set<XmlSchemaGenerator.Namespace> depends;
        private boolean selfReference;
        private final Set<ClassInfo<T, C>> classes;
        private final Set<EnumLeafInfo<T, C>> enums;
        private final Set<ArrayInfo<T, C>> arrays;
        private final MultiMap<String, AttributePropertyInfo<T, C>> attributeDecls;
        private final MultiMap<String, XmlSchemaGenerator.Namespace.ElementDeclaration> elementDecls;
        private Form attributeFormDefault;
        private Form elementFormDefault;
        private boolean useSwaRef;
        private boolean useMimeNs;
        private final Set<ClassInfo> written;
        
        public Namespace(String uri) {
            this.depends = new LinkedHashSet();
            this.classes = new LinkedHashSet();
            this.enums = new LinkedHashSet();
            this.arrays = new LinkedHashSet();
            this.attributeDecls = new MultiMap(null);
            this.elementDecls = new MultiMap(new XmlSchemaGenerator.Namespace.ElementWithType(true, XmlSchemaGenerator.this.anyType));
            this.written = new HashSet();
            this.uri = uri;
            assert !XmlSchemaGenerator.this.namespaces.containsKey(uri);
            XmlSchemaGenerator.this.namespaces.put(uri, this);
            return;
        }
        
        void resetWritten() {
            this.written.clear();
        }
        
        private void processForeignNamespaces(PropertyInfo<T, C> p, int processingDepth) {
            for (TypeInfo<T, C> t : p.ref()) {
                if (t instanceof ClassInfo && processingDepth > 0) {
                    List<PropertyInfo> l = ((ClassInfo)t).getProperties();
                    for (PropertyInfo subp : l) {
                        this.processForeignNamespaces(subp, --processingDepth);
                    }
                }
                if ((t instanceof Element)) {
                    this.addDependencyTo(((Element)t).getElementName());
                }
                if ((t instanceof NonElement)) {
                    this.addDependencyTo(((NonElement)t).getTypeName());
                }
                else {
                    continue;
                }
            }
        }
        
        private void addDependencyTo(@Nullable QName qname) {
            if (qname == null) {
                return;
            }
            String nsUri = qname.getNamespaceURI();
            if (nsUri.equals("http://www.w3.org/2001/XMLSchema")) {
                return;
            }
            if (nsUri.equals(this.uri)) {
                this.selfReference = true;
                return;
            }
            this.depends.add(XmlSchemaGenerator.access$1100(XmlSchemaGenerator.this, nsUri));
            return;
        }
        
        private void writeTo(Result result, Map<XmlSchemaGenerator.Namespace, String> systemIds) throws IOException {
            try {
                Schema schema = (Schema)TXW.create(Schema.class, ResultFactory.createSerializer(result));
                Map<String, String> xmlNs = XmlSchemaGenerator.this.types.getXmlNs(this.uri);
                for (Map.Entry<String, String> e : xmlNs.entrySet()) {
                    schema._namespace((String)e.getValue(), (String)e.getKey());
                }
                if (this.useSwaRef) {
                    schema._namespace("http://ws-i.org/profiles/basic/1.1/xsd", "swaRef");
                }
                if (this.useMimeNs) {
                    schema._namespace("http://www.w3.org/2005/05/xmlmime", "xmime");
                }
                (this.attributeFormDefault = Form.get(XmlSchemaGenerator.this.types.getAttributeFormDefault(this.uri))).declare("attributeFormDefault", schema);
                (this.elementFormDefault = Form.get(XmlSchemaGenerator.this.types.getElementFormDefault(this.uri))).declare("elementFormDefault", schema);
                if (!xmlNs.containsValue("http://www.w3.org/2001/XMLSchema") && !xmlNs.containsKey("xs")) {
                    schema._namespace("http://www.w3.org/2001/XMLSchema", "xs");
                }
                schema.version("1.0");
                if (this.uri.length() != 0) {
                    schema.targetNamespace(this.uri);
                }
                for (XmlSchemaGenerator.Namespace ns : this.depends) {
                    schema._namespace(ns.uri);
                }
                if (this.selfReference && this.uri.length() != 0) {
                    schema._namespace(this.uri, "tns");
                }
                schema._pcdata("\n");
                for (XmlSchemaGenerator.Namespace n : this.depends) {
                    Import imp = schema._import();
                    if (n.uri.length() != 0) {
                        imp.namespace(n.uri);
                    }
                    String refSystemId = (String)systemIds.get(n);
                    if (refSystemId != null && !refSystemId.equals("")) {
                        imp.schemaLocation(XmlSchemaGenerator.relativize(refSystemId, result.getSystemId()));
                    }
                    schema._pcdata("\n");
                }
                if (this.useSwaRef) {
                    schema._import().namespace("http://ws-i.org/profiles/basic/1.1/xsd").schemaLocation("http://ws-i.org/profiles/basic/1.1/swaref.xsd");
                }
                if (this.useMimeNs) {
                    schema._import().namespace("http://www.w3.org/2005/05/xmlmime").schemaLocation("http://www.w3.org/2005/05/xmlmime");
                }
                for (Map.Entry<String, XmlSchemaGenerator.Namespace.ElementDeclaration> e2 : this.elementDecls.entrySet()) {
                    ((XmlSchemaGenerator.Namespace.ElementDeclaration)e2.getValue()).writeTo((String)e2.getKey(), schema);
                    schema._pcdata("\n");
                }
                for (ClassInfo<T, C> c : this.classes) {
                    if (c.getTypeName() == null) {
                        continue;
                    }
                    if (this.uri.equals(c.getTypeName().getNamespaceURI())) {
                        this.writeClass(c, schema);
                    }
                    schema._pcdata("\n");
                }
                for (EnumLeafInfo<T, C> e3 : this.enums) {
                    if (e3.getTypeName() == null) {
                        continue;
                    }
                    if (this.uri.equals(e3.getTypeName().getNamespaceURI())) {
                        this.writeEnum(e3, schema);
                    }
                    schema._pcdata("\n");
                }
                for (ArrayInfo<T, C> a : this.arrays) {
                    this.writeArray(a, schema);
                    schema._pcdata("\n");
                }
                for (Map.Entry<String, AttributePropertyInfo<T, C>> e4 : this.attributeDecls.entrySet()) {
                    TopLevelAttribute a2 = schema.attribute();
                    a2.name((String)e4.getKey());
                    if (e4.getValue() == null) {
                        this.writeTypeRef(a2, XmlSchemaGenerator.this.stringType, "type");
                    }
                    else {
                        this.writeAttributeTypeRef((AttributePropertyInfo)e4.getValue(), a2);
                    }
                    schema._pcdata("\n");
                }
                schema.commit();
            }
            catch (TxwException e5) {
                XmlSchemaGenerator.logger.log(Level.INFO, e5.getMessage(), e5);
                throw new IOException(e5.getMessage());
            }
        }
        
        private void writeTypeRef(TypeHost th, NonElementRef<T, C> typeRef, String refAttName) {
            switch (typeRef.getSource().id()) {
                case ID: {
                    th._attribute(refAttName, new QName("http://www.w3.org/2001/XMLSchema", "ID"));
                    return;
                }
                case IDREF: {
                    th._attribute(refAttName, new QName("http://www.w3.org/2001/XMLSchema", "IDREF"));
                    return;
                }
                case NONE: {
                    MimeType mimeType = typeRef.getSource().getExpectedMimeType();
                    if (mimeType != null) {
                        th._attribute(new QName("http://www.w3.org/2005/05/xmlmime", "expectedContentTypes", "xmime"), mimeType.toString());
                    }
                    if (XmlSchemaGenerator.access$1500(XmlSchemaGenerator.this, typeRef)) {
                        th._attribute(refAttName, new QName("http://ws-i.org/profiles/basic/1.1/xsd", "swaRef", "ref"));
                        return;
                    }
                    if (typeRef.getSource().getSchemaType() != null) {
                        th._attribute(refAttName, typeRef.getSource().getSchemaType());
                        return;
                    }
                    this.writeTypeRef(th, typeRef.getTarget(), refAttName);
                    return;
                }
                default: {
                    throw new IllegalStateException();
                }
            }
        }
        
        private void writeTypeRef(TypeHost th, NonElement<T, C> type, String refAttName) {
            Element e = null;
            if ((type instanceof MaybeElement)) {
                MaybeElement me = (MaybeElement)type;
                boolean isElement = me.isElement();
                if (isElement) {
                    e = me.asElement();
                }
            }
            if ((type instanceof Element)) {
                e = (Element)type;
            }
            if (type.getTypeName() == null) {
                if (e != null && e.getElementName() != null) {
                    th.block();
                    if ((type instanceof ClassInfo)) {
                        this.writeClass((ClassInfo)type, th);
                    }
                    else {
                        this.writeEnum((EnumLeafInfo)type, (SimpleTypeHost)th);
                    }
                }
                else {
                    th.block();
                    if ((type instanceof ClassInfo)) {
                        if (XmlSchemaGenerator.this.collisionChecker.push((ClassInfo)type)) {
                            XmlSchemaGenerator.this.errorListener.warning(new SAXParseException(Messages.ANONYMOUS_TYPE_CYCLE.format(new Object[] { XmlSchemaGenerator.this.collisionChecker.getCycleString() }), null));
                        }
                        else {
                            this.writeClass((ClassInfo)type, th);
                        }
                        XmlSchemaGenerator.this.collisionChecker.pop();
                    }
                    else {
                        this.writeEnum((EnumLeafInfo)type, (SimpleTypeHost)th);
                    }
                }
            }
            else {
                th._attribute(refAttName, type.getTypeName());
            }
        }
        
        private void writeArray(ArrayInfo<T, C> a, Schema schema) {
            ComplexType ct = schema.complexType().name(a.getTypeName().getLocalPart());
            ct._final("#all");
            LocalElement le = ct.sequence().element().name("item");
            le.type(a.getItemType().getTypeName());
            le.minOccurs(0).maxOccurs("unbounded");
            le.nillable(true);
            ct.commit();
        }
        
        private void writeEnum(EnumLeafInfo<T, C> e, SimpleTypeHost th) {
            SimpleType st = th.simpleType();
            this.writeName(e, st);
            SimpleRestrictionModel base = st.restriction();
            this.writeTypeRef(base, e.getBaseType(), "base");
            for (EnumConstant c : e.getConstants()) {
                base.enumeration().value(c.getLexicalValue());
            }
            st.commit();
        }
        
        private void writeClass(ClassInfo<T, C> c, TypeHost parent) {
            if (this.written.contains(c)) {
                return;
            }
            this.written.add(c);
            if (this.containsValueProp(c)) {
                if (c.getProperties().size() == 1) {
                    ValuePropertyInfo<T, C> vp2 = (ValuePropertyInfo)c.getProperties().get(0);
                    SimpleType st = ((SimpleTypeHost)parent).simpleType();
                    this.writeName(c, st);
                    if (vp2.isCollection()) {
                        this.writeTypeRef(st.list(), vp2.getTarget(), "itemType");
                    }
                    else {
                        this.writeTypeRef(st.restriction(), vp2.getTarget(), "base");
                    }
                    return;
                }
                else {
                    ComplexType ct = ((ComplexTypeHost)parent).complexType();
                    this.writeName(c, ct);
                    if (c.isFinal()) {
                        ct._final("extension restriction");
                    }
                    SimpleExtension se2 = ct.simpleContent().extension();
                    se2.block();
                    for (PropertyInfo<T, C> p3 : c.getProperties()) {
                        switch (p3.kind()) {
                            case ATTRIBUTE: {
                                this.handleAttributeProp((AttributePropertyInfo)p3, se2);
                                continue;
                            }
                            case VALUE: {
                                TODO.checkSpec("what if vp.isCollection() == true?");
                                ValuePropertyInfo vp = (ValuePropertyInfo)p3;
                                se2.base(vp.getTarget().getTypeName());
                                continue;
                            }
                            default: {
                                assert false;
                                throw new IllegalStateException();
                            }
                        }
                    }
                    se2.commit();
                    TODO.schemaGenerator("figure out what to do if bc != null");
                    TODO.checkSpec("handle sec 8.9.5.2, bullet #4");
                    return;
                }
            }
            else {
                ComplexType ct = ((ComplexTypeHost)parent).complexType();
                this.writeName(c, ct);
                if (c.isFinal()) {
                    ct._final("extension restriction");
                }
                if (c.isAbstract()) {
                    ct._abstract(true);
                }
                AttrDecls contentModel = ct;
                TypeDefParticle contentModelOwner = ct;
                ClassInfo<T, C> bc = c.getBaseClass();
                if (bc != null) {
                    if (bc.hasValueProperty()) {
                        SimpleExtension se = contentModel = ct.simpleContent().extension();
                        contentModelOwner = null;
                        se.base(bc.getTypeName());
                    }
                    else {
                        ComplexExtension ce = contentModelOwner = (contentModel = ct.complexContent().extension());
                        ce.base(bc.getTypeName());
                    }
                }
                if (contentModelOwner != null) {
                    ArrayList<Tree> children = new ArrayList();
                    for (PropertyInfo<T, C> p : c.getProperties()) {
                        if (p instanceof ReferencePropertyInfo && ((ReferencePropertyInfo)p).isMixed()) {
                            ct.mixed(true);
                        }
                        Tree t = this.buildPropertyContentModel(p);
                        if (t != null) {
                            children.add(t);
                        }
                        else {
                            continue;
                        }
                    }
                    Tree top = Tree.makeGroup(c.isOrdered() ? GroupKind.SEQUENCE : GroupKind.ALL, children);
                    top.write(contentModelOwner);
                }
                for (PropertyInfo<T, C> p2 : c.getProperties()) {
                    if ((p2 instanceof AttributePropertyInfo)) {
                        this.handleAttributeProp((AttributePropertyInfo)p2, contentModel);
                    }
                    else {
                        continue;
                    }
                }
                if (c.hasAttributeWildcard()) {
                    contentModel.anyAttribute().namespace("##other").processContents("skip");
                }
                ct.commit();
                return;
            }
        }
        
        private void writeName(NonElement<T, C> c, TypedXmlWriter xw) {
            QName tn = c.getTypeName();
            if (tn != null) {
                xw._attribute("name", tn.getLocalPart());
            }
        }
        
        private boolean containsValueProp(ClassInfo<T, C> c) {
            for (PropertyInfo p : c.getProperties()) {
                if ((p instanceof ValuePropertyInfo)) {
                    return true;
                }
                continue;
            }
            return false;
        }
        
        private Tree buildPropertyContentModel(PropertyInfo<T, C> p) {
            switch (p.kind()) {
                case ELEMENT: {
                    return this.handleElementProp((ElementPropertyInfo)p);
                }
                case ATTRIBUTE: {
                    return null;
                }
                case REFERENCE: {
                    return this.handleReferenceProp((ReferencePropertyInfo)p);
                }
                case MAP: {
                    return this.handleMapProp((MapPropertyInfo)p);
                }
                case VALUE: {
                    assert false;
                    throw new IllegalStateException();
                }
                default: {
                    assert false;
                    throw new IllegalStateException();
                }
            }
        }
        
        private Tree handleElementProp(ElementPropertyInfo<T, C> ep) {
            if (ep.isValueList()) {
                return new Tree.Term() {
                    @Override
                    protected void write(ContentModelContainer parent, boolean isOptional, boolean repeated) {
                        TypeRef<T, C> t = (TypeRef)ep.getTypes().get(0);
                        LocalElement e = parent.element();
                        e.block();
                        QName tn = t.getTagName();
                        e.name(tn.getLocalPart());
                        com.sun.xml.internal.bind.v2.schemagen.xmlschema.List lst = e.simpleType().list();
                        XmlSchemaGenerator.Namespace.access$1800(XmlSchemaGenerator.Namespace.this, lst, t, "itemType");
                        XmlSchemaGenerator.Namespace.this.elementFormDefault.writeForm(e, tn);
                        this.writeOccurs(e, isOptional || !ep.isRequired(), repeated);
                    }
                };
            }
            ArrayList<Tree> children = new ArrayList();
            for (TypeRef<T, C> t : ep.getTypes()) {
                children.add(new Tree.Term() {
                    @Override
                    protected void write(ContentModelContainer parent, boolean isOptional, boolean repeated) {
                        LocalElement e = parent.element();
                        QName tn = t.getTagName();
                        PropertyInfo propInfo = t.getSource();
                        TypeInfo parentInfo = (propInfo == null) ? null : propInfo.parent();
                        if (XmlSchemaGenerator.Namespace.access$2000(XmlSchemaGenerator.Namespace.this, t, tn, parentInfo)) {
                            if (!t.getTarget().isSimpleType() && t.getTarget() instanceof ClassInfo && XmlSchemaGenerator.this.collisionChecker.findDuplicate((ClassInfo)t.getTarget())) {
                                e.ref(new QName(XmlSchemaGenerator.Namespace.this.uri, tn.getLocalPart()));
                            }
                            else {
                                QName elemName = null;
                                if ((t.getTarget() instanceof Element)) {
                                    Element te = (Element)t.getTarget();
                                    elemName = te.getElementName();
                                }
                                Collection<TypeInfo> refs = propInfo.ref();
                                TypeInfo ti;
                                if (refs != null && !refs.isEmpty() && elemName != null && ((ti = (TypeInfo)refs.iterator().next()) == null || ti instanceof ClassInfoImpl)) {
                                    ClassInfoImpl cImpl = (ClassInfoImpl)ti;
                                    if (cImpl != null && cImpl.getElementName() != null) {
                                        e.ref(new QName(cImpl.getElementName().getNamespaceURI(), tn.getLocalPart()));
                                    }
                                    else {
                                        e.ref(new QName("", tn.getLocalPart()));
                                    }
                                }
                                else {
                                    e.ref(tn);
                                }
                            }
                        }
                        else {
                            e.name(tn.getLocalPart());
                            XmlSchemaGenerator.Namespace.access$1800(XmlSchemaGenerator.Namespace.this, e, t, "type");
                            XmlSchemaGenerator.Namespace.this.elementFormDefault.writeForm(e, tn);
                        }
                        if (t.isNillable()) {
                            e.nillable(true);
                        }
                        if (t.getDefaultValue() != null) {
                            e._default(t.getDefaultValue());
                        }
                        this.writeOccurs(e, isOptional, repeated);
                    }
                });
            }
            Tree choice = Tree.makeGroup(GroupKind.CHOICE, children).makeOptional(!ep.isRequired()).makeRepeated(ep.isCollection());
            QName ename = ep.getXmlName();
            if (ename != null) {
                return new Tree.Term() {
                    @Override
                    protected void write(ContentModelContainer parent, boolean isOptional, boolean repeated) {
                        LocalElement e = parent.element();
                        if (ename.getNamespaceURI().length() > 0 && !ename.getNamespaceURI().equals(XmlSchemaGenerator.Namespace.this.uri)) {
                            e.ref(new QName(ename.getNamespaceURI(), ename.getLocalPart()));
                            return;
                        }
                        e.name(ename.getLocalPart());
                        XmlSchemaGenerator.Namespace.this.elementFormDefault.writeForm(e, ename);
                        if (ep.isCollectionNillable()) {
                            e.nillable(true);
                        }
                        this.writeOccurs(e, !ep.isCollectionRequired(), repeated);
                        ComplexType p = e.complexType();
                        choice.write(p);
                        return;
                    }
                };
            }
            return choice;
        }
        
        private boolean canBeDirectElementRef(TypeRef<T, C> t, QName tn, TypeInfo parentInfo) {
            Element te = null;
            ClassInfo ci = null;
            QName targetTagName = null;
            if (t.isNillable() || t.getDefaultValue() != null) {
                return false;
            }
            if ((t.getTarget() instanceof Element)) {
                te = (Element)t.getTarget();
                targetTagName = te.getElementName();
                if ((te instanceof ClassInfo)) {
                    ci = (ClassInfo)te;
                }
            }
            String nsUri = tn.getNamespaceURI();
            return (!nsUri.equals(this.uri) && nsUri.length() > 0 && (!(parentInfo instanceof ClassInfo) || ((ClassInfo)parentInfo).getTypeName() != null)) || (ci != null && targetTagName != null && te.getScope() == null && targetTagName.getNamespaceURI() == null && targetTagName.equals(tn)) || (te != null && targetTagName != null && targetTagName.equals(tn));
        }
        
        private void handleAttributeProp(AttributePropertyInfo<T, C> ap, AttrDecls attr) {
            LocalAttribute localAttribute = attr.attribute();
            String attrURI = ap.getXmlName().getNamespaceURI();
            if (attrURI.equals("")) {
                localAttribute.name(ap.getXmlName().getLocalPart());
                this.writeAttributeTypeRef(ap, localAttribute);
                this.attributeFormDefault.writeForm(localAttribute, ap.getXmlName());
            }
            else {
                localAttribute.ref(ap.getXmlName());
            }
            if (ap.isRequired()) {
                localAttribute.use("required");
            }
        }
        
        private void writeAttributeTypeRef(AttributePropertyInfo<T, C> ap, AttributeType a) {
            if (ap.isCollection()) {
                this.writeTypeRef(a.simpleType().list(), ap, "itemType");
            }
            else {
                this.writeTypeRef(a, ap, "type");
            }
        }
        
        private Tree handleReferenceProp(ReferencePropertyInfo<T, C> rp) {
            ArrayList<Tree> children = new ArrayList();
            for (Element<T, C> e : rp.getElements()) {
                children.add(new Tree.Term() {
                    @Override
                    protected void write(ContentModelContainer parent, boolean isOptional, boolean repeated) {
                        LocalElement eref = parent.element();
                        boolean local = false;
                        QName en = e.getElementName();
                        if (e.getScope() != null) {
                            boolean qualified = en.getNamespaceURI().equals(XmlSchemaGenerator.Namespace.this.uri);
                            boolean unqualified = en.getNamespaceURI().equals("");
                            if (qualified || unqualified) {
                                if (unqualified) {
                                    if (XmlSchemaGenerator.Namespace.this.elementFormDefault.isEffectivelyQualified) {
                                        eref.form("unqualified");
                                    }
                                }
                                else if (!XmlSchemaGenerator.Namespace.this.elementFormDefault.isEffectivelyQualified) {
                                    eref.form("qualified");
                                }
                                local = true;
                                eref.name(en.getLocalPart());
                                if ((e instanceof ClassInfo)) {
                                    XmlSchemaGenerator.Namespace.access$2100(XmlSchemaGenerator.Namespace.this, eref, (ClassInfo)e, "type");
                                }
                                else {
                                    XmlSchemaGenerator.Namespace.access$2100(XmlSchemaGenerator.Namespace.this, eref, ((ElementInfo)e).getContentType(), "type");
                                }
                            }
                        }
                        if (!local) {
                            eref.ref(en);
                        }
                        this.writeOccurs(eref, isOptional, repeated);
                    }
                });
            }
            WildcardMode wc = rp.getWildcard();
            if (wc != null) {
                children.add(new Tree.Term() {
                    @Override
                    protected void write(ContentModelContainer parent, boolean isOptional, boolean repeated) {
                        Any any = parent.any();
                        String pcmode = XmlSchemaGenerator.access$2200(wc);
                        if (pcmode != null) {
                            any.processContents(pcmode);
                        }
                        any.namespace("##other");
                        this.writeOccurs(any, isOptional, repeated);
                    }
                });
            }
            Tree choice = Tree.makeGroup(GroupKind.CHOICE, children).makeRepeated(rp.isCollection()).makeOptional(!rp.isRequired());
            QName ename = rp.getXmlName();
            if (ename != null) {
                return new Tree.Term() {
                    @Override
                    protected void write(ContentModelContainer parent, boolean isOptional, boolean repeated) {
                        LocalElement e = parent.element().name(ename.getLocalPart());
                        XmlSchemaGenerator.Namespace.this.elementFormDefault.writeForm(e, ename);
                        if (rp.isCollectionNillable()) {
                            e.nillable(true);
                        }
                        this.writeOccurs(e, true, repeated);
                        ComplexType p = e.complexType();
                        choice.write(p);
                    }
                };
            }
            return choice;
        }
        
        private Tree handleMapProp(MapPropertyInfo<T, C> mp) {
            return new Tree.Term() {
                @Override
                protected void write(ContentModelContainer parent, boolean isOptional, boolean repeated) {
                    QName ename = mp.getXmlName();
                    LocalElement e = parent.element();
                    XmlSchemaGenerator.Namespace.this.elementFormDefault.writeForm(e, ename);
                    if (mp.isCollectionNillable()) {
                        e.nillable(true);
                    }
                    e = e.name(ename.getLocalPart());
                    this.writeOccurs(e, isOptional, repeated);
                    ComplexType p = e.complexType();
                    e = p.sequence().element();
                    e.name("entry").minOccurs(0).maxOccurs("unbounded");
                    ExplicitGroup seq = e.complexType().sequence();
                    XmlSchemaGenerator.Namespace.access$2300(XmlSchemaGenerator.Namespace.this, seq, "key", mp.getKeyType());
                    XmlSchemaGenerator.Namespace.access$2300(XmlSchemaGenerator.Namespace.this, seq, "value", mp.getValueType());
                }
            };
        }
        
        private void writeKeyOrValue(ExplicitGroup seq, String tagName, NonElement<T, C> typeRef) {
            LocalElement key = seq.element().name(tagName);
            key.minOccurs(0);
            this.writeTypeRef(key, typeRef, "type");
        }
        
        public void addGlobalAttribute(AttributePropertyInfo<T, C> ap) {
            this.attributeDecls.put(ap.getXmlName().getLocalPart(), ap);
            this.addDependencyTo(ap.getTarget().getTypeName());
        }
        
        public void addGlobalElement(TypeRef<T, C> tref) {
            this.elementDecls.put(tref.getTagName().getLocalPart(), new XmlSchemaGenerator.Namespace.ElementWithType(false, tref.getTarget()));
            this.addDependencyTo(tref.getTarget().getTypeName());
        }
        
        @Override
        public String toString() {
            StringBuilder buf = new StringBuilder();
            buf.append("[classes=").append(this.classes);
            buf.append(",elementDecls=").append(this.elementDecls);
            buf.append(",enums=").append(this.enums);
            buf.append("]");
            return super.toString();
        }
        
        static void access$100(XmlSchemaGenerator.Namespace x0, QName x1) {
            x0.addDependencyTo(x1);
        }
        
        static void access$200(XmlSchemaGenerator.Namespace x0, PropertyInfo x1, int x2) {
            x0.processForeignNamespaces(x1, x2);
        }
        
        static void access$800(XmlSchemaGenerator.Namespace x0, Result x1, Map x2) throws IOException {
            x0.writeTo(x1, x2);
        }
        
        static void access$1800(XmlSchemaGenerator.Namespace x0, TypeHost x1, NonElementRef x2, String x3) {
            x0.writeTypeRef(x1, x2, x3);
        }
        
        static boolean access$2000(XmlSchemaGenerator.Namespace x0, TypeRef x1, QName x2, TypeInfo x3) {
            return x0.canBeDirectElementRef(x1, x2, x3);
        }
        
        static void access$2100(XmlSchemaGenerator.Namespace x0, TypeHost x1, NonElement x2, String x3) {
            x0.writeTypeRef(x1, x2, x3);
        }
        
        static void access$2300(XmlSchemaGenerator.Namespace x0, ExplicitGroup x1, String x2, NonElement x3) {
            x0.writeKeyOrValue(x1, x2, x3);
        }
        
        abstract class ElementDeclaration
        {
            @Override
            public abstract boolean equals(Object p0);
            
            @Override
            public abstract int hashCode();
            
            public abstract void writeTo(String p0, Schema p1);
        }
        
        class ElementWithType extends XmlSchemaGenerator.Namespace.ElementDeclaration
        {
            private final boolean nillable;
            private final NonElement<T, C> type;
            
            public ElementWithType(boolean nillable, NonElement<T, C> type) {
                this.type = type;
                this.nillable = nillable;
            }
            
            @Override
            public void writeTo(String localName, Schema schema) {
                TopLevelElement e = schema.element().name(localName);
                if (this.nillable) {
                    e.nillable(true);
                }
                if (this.type != null) {
                    XmlSchemaGenerator.Namespace.access$2100(XmlSchemaGenerator.Namespace.this, e, this.type, "type");
                }
                else {
                    e.complexType();
                }
                e.commit();
            }
            
            @Override
            public boolean equals(Object o) {
                if (this == o) {
                    return true;
                }
                if (o == null || this.getClass() != o.getClass()) {
                    return false;
                }
                XmlSchemaGenerator.Namespace.ElementWithType that = (XmlSchemaGenerator.Namespace.ElementWithType)o;
                return this.type.equals(that.type);
            }
            
            @Override
            public int hashCode() {
                return this.type.hashCode();
            }
        }
    }
}
